#!/usr/bin/python

##########
# aggregates labeling stats
##########
import os, string, decimal, glob, re, subprocess, shlex, sys, datetime, ntpath, time, random
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../largescale/"))
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../experiment/"))
import mydb, expdb

Kmax = 50; # vk: const, maximal number of models
globThreshold = 150;	# vk: const, global energy threshold
eFrac = 2;			# vk: const
eMin = 20;
nModelsToLearn = 200; # number of models to learn template variations
energyThreshToSpawn = 75; # same as in TMatchEnergy

def groupByParts(modelList):
	groupedList = [];
	flags2list = dict();
	modelList.sort(key=lambda x: x.energy, reverse=False);
	eThresh = min(max(modelList[1].energy, eMin)*eFrac, globThreshold);
	for mInst in modelList:
		include = mInst.energy < eThresh; # include if energy is small enough
		if include:
			if mInst.partFlags in flags2list:
				gid = flags2list[mInst.partFlags];
			else:
				gid = len(groupedList);
				flags2list[mInst.partFlags] = gid;
				groupedList.append([]);
			if len(groupedList[gid]) < Kmax:
				groupedList[gid].append(mInst);

	print("Flag   # models");
	for mList in groupedList:
		print(mList[0].partFlags+": "+str(len(mList))+" e.g. "+str(mList[0].mname)+", "+str(mList[0].energy));

	return groupedList;

# INPUT: 	all models
# OUTPUT:	models to label
inListName2File = dict();
outListName2File = dict();

iter = 0;
mode = "none";

if len(sys.argv)>=1:
	i = 1
	while i<len(sys.argv):
		if (sys.argv[i][0:7]=="-inList"):
			lName = sys.argv[i][7:];
			i=i+1;
			inListName2File[lName] = sys.argv[i];
		elif (sys.argv[i][0:8]=="-outList"):
			lName = sys.argv[i][8:];
			i=i+1;
			outListName2File[lName] = sys.argv[i];
		elif (sys.argv[i]=="-iter"):
			i=i+1;
			iter = int(sys.argv[i]);
		elif (sys.argv[i]=="-mode"):
			i=i+1;
			mode = sys.argv[i];
		else:
			print("[WARNING] Cannot understand command: "+sys.argv[i]+" in selectSubset.py");
		i = i+1;
else:
	print("Usage:");
	print("   python "+sys.argv[0]+" [options]");
	print("   OPTIONS:");
	print("       -inList[Success, Remaining, Excluded, ToLabel, Labeled, ToDeform, Deformed, Clustered] modellist.txt");
	print("       -outList[ToLabel, ToDeform, Excluded, Success, Remaining, Excluded] modellist.txt");
	print("       -iter iterID");
	print("       -mode [SelectToLabel, SelectToDeform, UpdNextLevel]");
	exit();

start = time.clock();

if mode=="SelectToLabel":	# Select which models to label
	if not "Remaining" in inListName2File or not "ToLabel" in outListName2File:
		print("[ERROR] Missing label type: for mode="+mode);
		exit();
	mList = expdb.readModelList(inListName2File["Remaining"]);
	if iter==0:
		expdb.writeModelList(mList, outListName2File["ToLabel"], False);
	else:
		outFile = inListName2File["Remaining"];
		updFile = outFile[0:-4]+"_updSeeds.txt";

#		if os.path.isfile(updFile):	# template was updated
#			print("Template Was not updated");
#			nElements = min(Kmax, len(mList));	# vk: const
#			#random.shuffle(mList);
#			mList = mList[0:nElements];
#			expdb.writeModelList(mList, outListName2File["ToLabel"], False);
#		else:
		print("Template Was Updated");
		nElements = min(2 * Kmax, len(mList));	# vk: const
		nRand = 2 * Kmax;
		if nRand>len(mList[nElements:]):
			nRand = len(mList[nElements:]);
		mList = mList[0:nElements] + random.sample(mList[nElements:], nRand);
		expdb.writeModelList(mList, outListName2File["ToLabel"], False);
elif mode=="SelectToDeform":	# Select which models to include in Working Set
	if not "Labeled" in inListName2File or not "ToDeform" in outListName2File:
		print("[ERROR] Missing label type: for mode="+mode);
		exit();

	fullList = [];

	readList = expdb.readModelList(inListName2File["Labeled"]);
	toGroupList = [];
	for mInst in readList:
		include = mInst.ntmplt < nModelsToLearn or mInst.deformDist > energyThreshToSpawn;  # include if useful for update
		include = include and mInst.energy < globThreshold;
		if include: # deform, try to learn...
			toGroupList.append(mInst);
	groupedList = groupByParts(toGroupList);

	for mList in groupedList: # find models to deform for each part config (treat separately)
		mList.sort(key=lambda x: x.energy, reverse=False);
		if (len(mList)>1):
			eThresh = min(max(mList[1].energy, eMin)*eFrac, globThreshold);
			for i in range(0, min(Kmax, len(mList))):
				if (mList[i].energy < eThresh and mList[i].energy >= 0):
					fullList.append(mList[i]);
	print("Groups: "+str(len(groupedList))+" NModels: "+str(len(fullList)));
	if (len(fullList)>0):
		expdb.writeModelList(fullList, outListName2File["ToDeform"], False);
elif mode=="UpdNextLevel":	# Update Success / Remaining / Excluded lists
	validInput = True;
	#validInput = validInput and "Success" in inListName2File;
	#validInput = validInput and "Excluded" in inListName2File;
	validInput = validInput and "Remaining" in inListName2File;
	validInput = validInput and "Clustered" in inListName2File;
	validInput = validInput and "Deformed" in inListName2File;
	validInput = validInput and "Labeled" in inListName2File;	
	validOutput = True;
	validOutput = validOutput and "Success" in outListName2File;
	validOutput = validOutput and "Remaining" in outListName2File;
	validOutput = validOutput and "Excluded" in outListName2File;
	if not validInput or not validOutput:
		print("[ERROR] Missing label type: for mode="+mode);
		exit();
	mSuccesss = expdb.readModelList(inListName2File["Success"]);
	mRemaining = expdb.readModelList(inListName2File["Remaining"]);
	mExcluded = expdb.readModelList(inListName2File["Excluded"]);
	mDeformed = expdb.readModelList(inListName2File["Deformed"]);
	mClustered = expdb.readModelList(inListName2File["Clustered"]);
	mLabeled = expdb.readModelList(inListName2File["Labeled"]);

	updFile = inListName2File["Clustered"];
	updFile = updFile[0:-4]+"_updSeeds.txt";
	if os.path.isfile(updFile):
		outFile = outListName2File["Remaining"];
		os.system("cp "+updFile+" "+outFile[0:-4]+"_updSeeds.txt");

	outSuccess = mSuccesss;
	outRemaining = [];
	outExcluded = mExcluded;

	for m2 in mRemaining:
		wasDeformed = False;
		wasClustered = False;
		wasLabeled = False;
		noneDeformed = len(mDeformed)==0;
		for m in mLabeled:
			if (m.mname == m2.mname):
				wasLabeled = True;
				break;
		for m in mDeformed:
			if (m.mname == m2.mname):
				wasDeformed = True;
				break;
		for m in mClustered:
			if (m.mname == m2.mname):
				wasClustered = True;
				break;
		mInst = m2;
		include = mInst.ntmplt < nModelsToLearn or mInst.deformDist > energyThreshToSpawn;  # include if useful for update
		include = include and mInst.energy < globThreshold;
		if (wasClustered):		# deformed, clustered => success
			outSuccess.append(m2);
		elif ((wasDeformed or (wasLabeled and noneDeformed)) and iter>1 and len(mRemaining)<Kmax): #
			# (deformed, not clustered) OR (labeled, deformed=empty) => exclude
			# only for small sets
			outExcluded.append(m2);
		elif include:					# not deformed, not clustered => revisit...
			outRemaining.append(m2);
	# none labeled - something is wrong... - stop
	if (len(mLabeled)==0):
		outExcluded = outExcluded + outRemaining;
		outRemaining = [];

	expdb.writeModelList(outSuccess, outListName2File["Success"], False);
	expdb.writeModelList(outRemaining, outListName2File["Remaining"], False);
	expdb.writeModelList(outExcluded, outListName2File["Excluded"], False);
else:
	print("[ERROR] Invalid subset selection mode");
	exit();

if mydb.WebspaceExists() and len(outListName2File)>0:
	elapsed = (time.clock() - start);
	key, value = outListName2File.popitem();
	with open(ntpath.dirname(value)+"/selectTime.txt", "a") as timeFile:
		timeFile.write(mode+" "+str(elapsed));




